/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.bndmgt;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.osgi.framework.*;
import org.osgi.service.startlevel.StartLevel;
import org.osgi.util.tracker.ServiceTracker;

import java.util.Dictionary;
import java.util.Enumeration;

/**
 * This agent is responsible for extracting the start level custom header from the bundle manifest
 * headers and setting that start level to the bundle through the start level service. Further
 * when a new bundle comes to the install phase of the bundle lifecycle it will also be picked by
 * the agent.</p>
 * <p/>
 * <p>This implements the {@link org.osgi.framework.SynchronousBundleListener} and hence framework
 * will wait till this listener returns guaranteeing that all the bundles which are installed have
 * the required strat levels set.
 *
 * @see org.osgi.framework.SynchronousBundleListener
 */
public class BundleManagementAgent implements BundleListener, FrameworkListener {

    /**
     * Used to store the Log variable reference
     */
    private static final Log log = LogFactory.getLog(BundleManagementAgent.class);

    /**
     * Keeps the reference to the BundleContext
     */
    BundleContext bundleContext;

    /**
     * Keep track of the ServiceTracker to access the start level service
     */
    ServiceTracker startLevelServiceTracker;

    /**
     * Keeps the current framework start level because at the framework initialization
     * time this is going to be different from the actual framework start level
     */
    private int currentFrameworkStartLevel;

    /**
     * Lock variable to synchronize the starting of the bundle.
     *
     * We cannot use the {@link org.osgi.framework.SynchronousBundleListener} because it
     * will lead to a race condition according to the logic of the agent
     */
    private final Object lock = new Object();

    /**
     * This is to hold off the bundle listener in busy wait mode till the initial starting
     * of the bundles is completed
     */
    private boolean initialized = false;

    /**
     * Singleton instance reference
     */
    private static BundleManagementAgent instance;

    /**
     * Constructor for the agent using the {@link org.osgi.framework.BundleContext}
     *
     * @param bundleContext information about the bundle who is constructing the agent and the
     *                      information on the OSGi run time
     * @throws CarbonBundleException if the StartLevel service cannot be found
     */
    private BundleManagementAgent(BundleContext bundleContext) throws CarbonBundleException {

        if (log.isDebugEnabled()) {
            log.debug("Initializing the Bundle Management Agent...");
        }

        this.bundleContext = bundleContext;
        // uses the service tracker to access the start level service
        startLevelServiceTracker = new ServiceTracker(
                bundleContext, StartLevel.class.getName(), null);
        startLevelServiceTracker.open();
        StartLevel startLevelService = (StartLevel) startLevelServiceTracker.getService();

        // if the start level service is available then use it to set the
        // management agent start level
        if (startLevelService != null) {
            startLevelService.setBundleStartLevel(bundleContext.getBundle(),
                    BundleManagementConstants.BUNDLE_MANAGEMENT_AGENT_START_LEVEL);
        } else {
            handleException("Couldn't retrieve the StartLevel service");
        }

        // closes the reference to the start level service
        startLevelServiceTracker.close();

        if (log.isDebugEnabled()) {
            log.debug("Successfully initialized the Bundle Management Agent...");
        }
    }

    /**
     * BundleManagementAgent has to be a singleton because it keeps track of the current
     * start level of the framework. This is sort of a extra precaution, but better keep
     * as it is, because this is a piece of very sensitive code for OSGi framework to
     * properly start.
     *
     * @param bundleContext context of the management agent
     * @return singleton instance of the <code>BundleManagementAgent</code>
     * @throws CarbonBundleException if the StartLevel service cannot be found
     */
    public static BundleManagementAgent getInstance(BundleContext bundleContext)
            throws CarbonBundleException {

        if (instance == null) {
            instance = new BundleManagementAgent(bundleContext);
        }
        return instance;
    }

    /**
     * This will be called when there is a new bundle installed and it will be assigned with the
     * correct start level from the manifest header.
     *
     * @param event describing the bundle information
     */
    public void bundleChanged(BundleEvent event) {

        // if this is a INSALLED bundle event
        if (event.getType() == BundleEvent.INSTALLED || event.getType() == BundleEvent.RESOLVED) {

            if (log.isDebugEnabled()) {
                log.debug("Bundle with symbolic name : " + event.getBundle().getSymbolicName()
                        + " has been installed");
            }

            // Busy wait till the initial startup of the existing bundles complete
            while (!initialized) {
                try {
                    Thread.sleep(100);
                } catch (InterruptedException ignore) {}
            }

            // IMPORTANT : extra care has to be taken in handling this piece of code,
            // we cannot use the SynchronousBundleListener here becasue it leads to a race
            // condition with the above is initialized busy wait check.
            synchronized (lock) {

                // use the service tracker to get the start level service and set the start level
                // as appropriate to the bundle
                startLevelServiceTracker.open();
                StartLevel startLevelService = (StartLevel) startLevelServiceTracker.getService();

                try {
                    // sets the start level to the just installed bundle and
                    // mark it for starting
                    int frameworkStartLevel = startLevelService.getStartLevel();

                    // incase if the framework has increased the framework start level adjust
                    // our selves to the framework, without back tracking the start level
                    if (currentFrameworkStartLevel < frameworkStartLevel) {
                        currentFrameworkStartLevel = frameworkStartLevel;
                    }
                    if (log.isDebugEnabled()) {
                        log.debug("Framework Start Level is : " + currentFrameworkStartLevel);
                    }
                    
                    int newFrameworkStartLevel = setStartLevel(
                            event.getBundle(), currentFrameworkStartLevel, startLevelService);

                    // if the start level of the just installed bundle is higher than the framework
                    // start level then, increase the framework start level
                    if (newFrameworkStartLevel > currentFrameworkStartLevel) {
                        
                        if (log.isDebugEnabled()) {
                            log.debug("Framework Start Level increased from : "
                                    + currentFrameworkStartLevel
                                    + " to : " + newFrameworkStartLevel);
                        }
                        currentFrameworkStartLevel = newFrameworkStartLevel;
                        startLevelService.setStartLevel(currentFrameworkStartLevel);
                    }

                } catch (CarbonBundleException e) {
                    log.error("Couldn't set the start level of the bundle : "
                            + event.getBundle().getSymbolicName(), e);
                }
                
                // closes the reference to the StartLevel service
                startLevelServiceTracker.close();
            }
        }
    }

    /**
     * Listen for the framework events to trigger the initializing of the bundles. We cannot
     * listen for the FRAMEWORK_STARTED event (if it would have been possible that would be
     * the ideal) because when we register the listener, it has already been thrown.
     *
     * Therefore we happily listen the STARTLEVEL_CHANGED event and look for the required value
     * for me to get started my work. :-)
     *
     * @param frameworkEvent event of the framework
     */
    public void frameworkEvent(FrameworkEvent frameworkEvent) {

        if (frameworkEvent.getType() == FrameworkEvent.STARTLEVEL_CHANGED) {
            // Open the start level service tracker and get the start level service
            startLevelServiceTracker.open();
            StartLevel startLevel = (StartLevel) startLevelServiceTracker.getService();

            // if this event is what we are looking for (i.e. the
            // required start level changed event)
            if (startLevel.getStartLevel()
                    == BundleManagementConstants.BUNDLE_MANAGEMENT_ACTIVATION_START_LEVEL) {
                
                try {

                    if (log.isDebugEnabled()) {
                        log.debug("Bundle Management Agent activated ... Starting the rest " +
                                "of the bundles according to there start levels");
                    }
                    // process the initial start levels
                    setInitStartLevels();
                    // mark the management agent as intialized and let the bundle listeners
                    // to be activated
                    initialized = true;
                    
                } catch (CarbonBundleException e) {
                    String message = "Unable to set the initial start levels";
                    log.fatal(message, e);
                    throw new RuntimeException(message, e);
                }
            }
        }
    }

    /**
     * This will set the initial start levels for all the bundles which are containing the
     * <code>WSO2-Bundle-StartLevel</code> header in the MANIFEST. If a particular bundle is having
     * a start level which is higher than the framework start level then the framework start level
     * will be increased to that (start level + 1).</p>
     * <p/>
     * <p>Bundles which does not contain any information about its start level will get the default
     * start level of the framework. In the process of adding the start levels to the bundles this
     * will call the start method of the bundle, which just marks the bundle to be started but the
     * bundle will actually be started by the framework according to its start levels.
     *
     * @return the final framework start level as an int
     * @throws CarbonBundleException incase if a particular bundle cannot be started
     */
    public int setInitStartLevels() throws CarbonBundleException {

        if (log.isDebugEnabled()) {
            log.debug("Setting the initial start levels to the installed bundles");
        }

        // Retrievs all the bundles
        Bundle[] bundles = bundleContext.getBundles();

        // gets the StartLevel service reference from the tracker
        startLevelServiceTracker.open();
        StartLevel startLevelService = (StartLevel) startLevelServiceTracker.getService();
        int frameworkStartLevel = startLevelService.getStartLevel();
        if (log.isDebugEnabled()) {
            log.debug("Initial start level of the framework is : " + frameworkStartLevel);
        }

        for (Bundle b : bundles) {
            // calls the setStartLevel whihc contains the logic to set the start level
            // if there is a one for all the bundles. This method will return the new start
            // level of the framework
            frameworkStartLevel = setStartLevel(b, frameworkStartLevel, startLevelService);
        }

        if (log.isDebugEnabled()) {
            log.debug("Setting the final start level of the framework as : " + frameworkStartLevel);
        }

        // set the current framework start level to the instance variable for the reference
        // of the bundle listener
        currentFrameworkStartLevel = frameworkStartLevel;
        // finaly modify the framework start level to the new value
        startLevelService.setStartLevel(currentFrameworkStartLevel);

        // closes the reference to the start level service
        startLevelServiceTracker.close();
        
        return frameworkStartLevel;
    }

    /**
     * This will set the start level of a particular bundle if there is the
     * <code>WSO2-Bundle-StartLevel</code> manifest header with a valid positive integer.</p>
     * <p/>
     * <p>Reads the manifest header and sests the start level using the start level service. Then
     * this looks for the framework start level and if the start level of the bundle is greater
     * than the current start level of the framework then it increases the start level of the
     * framework to (bundle start level + 1)
     *
     * @param bundle              of which the start level has to be set
     * @param frameworkStartLevel current framework start level
     * @param startLevelService   start level service reference to be used to set the start level
     * @return modified new start level of the framework
     * @throws CarbonBundleException incase if the bundle cannot be started
     */
    public int setStartLevel(Bundle bundle, int frameworkStartLevel,
                             StartLevel startLevelService) throws CarbonBundleException {

        // if the bundle is in INSTALLED or RESOLVED states
        if (bundle.getState() == Bundle.INSTALLED || bundle.getState() == Bundle.RESOLVED) {

            // reads the manifest headers and tries to extract the start level
            int startLevel = getStartLevel(bundle);
            // if there is a non negative start level defined in the manifest header
            if (startLevel > 1) {

                if (log.isDebugEnabled()) {
                    log.debug("Setting the start level : " + startLevel
                            + " for the bundle : " + bundle.getSymbolicName());
                }
                // sets the bundle start level to the bundle
                startLevelService.setBundleStartLevel(bundle, startLevel);

                try {
                    if (log.isDebugEnabled()) {
                        log.debug("Marking the bundle : " + bundle.getSymbolicName()
                                + " to be started by the framework");
                    }
                    // mark the bundle to be started
                    bundle.start();
                } catch (BundleException e) {
                    handleException("Couldn't start the bundle : " + bundle.getSymbolicName(), e);
                }
            } else {
                // Start Level manifest header is not specified here. But need to check whether
                // this is a fragment bundle.
                Dictionary headers = bundle.getHeaders();
                if (headers.get(Constants.FRAGMENT_HOST) == null) {
                    try {
                        if (log.isDebugEnabled()) {
                            log.debug("Marking the bundle : " + bundle.getSymbolicName()
                                    + " to be started by the framework with the default start level");
                        }
                        bundle.start();
                    } catch (BundleException e) {
                        log.error("Couldn't start the bundle : " + bundle.getSymbolicName()
                                + " with the default bundle start level", e);
                    }
                }
            }

            // if the bundle start level is greater than the framework start level increase the
            // framework start level to (bundle start level + 1)
            if (startLevel > frameworkStartLevel) {
                if (log.isDebugEnabled()) {
                    log.debug("Framework satrt level is marked to be increased from : "
                            + frameworkStartLevel + " to : " + (startLevel + 1));
                }
                frameworkStartLevel = startLevel + 1;
            }
        }

        return frameworkStartLevel;
    }

    /**
     * This reads the bundle manifest headers and tries to find the
     * <code>WSO2-Bundle-StartLevel</code> header. and returns the start level.
     *
     * @param bundle from which the start level is extracted out
     * @return bundle start level as an non negative integer if found -1 if not found
     */
    private int getStartLevel(Bundle bundle) {

        // retrieves all the manifest headers
        Dictionary dictionary = bundle.getHeaders();
        if (dictionary != null) {

            // Iterate through the headers and find the start level custom header
            for (Enumeration e = dictionary.keys(); e.hasMoreElements();) {

                String headerKey = (String) e.nextElement();
                if (BundleManagementConstants.START_LEVEL_MANIFEST_HEADER.equals(headerKey)) {

                    // retireve the header value
                    String headerValue = (String) dictionary.get(headerKey);
                    if (log.isDebugEnabled()) {
                        log.debug("Start level header with name : " + headerKey
                                + " has the value : " + headerValue + " in bundle : "
                                + bundle.getSymbolicName());
                    }

                    try {
                        // tries to get the integer representaiton of the header value
                        return Integer.valueOf(headerValue);
                    } catch (NumberFormatException nfe) {
                        log.error("Start Level has to be a non negative integer; found : "
                                + headerValue);
                    }
                }
            }
        }

        // if there is no Start Level information in the manifest headders return -1
        return -1;
    }

    // Utility methods for exception handling
    private void handleException(String message) throws CarbonBundleException {
        log.error(message);
        throw new CarbonBundleException(message);
    }

    private void handleException(String message, Throwable exception) throws CarbonBundleException {
        log.error(message, exception);
        throw new CarbonBundleException(message, exception);
    }
}
