/*
 * Copyright 2004,2005 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.bridge;

import static org.wso2.carbon.bridge.FrameworlLauncherConstants.*;

import java.io.*;
import java.util.ArrayList;
import java.util.List;
import java.util.jar.Attributes;
import java.util.jar.JarOutputStream;
import java.util.jar.Manifest;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

/**
 * This class will read the extension folder and create extension bundles.
 * User need to drop the classic jars that needed to be extended in ${carbon.home}/extensions
 * folder.
 */
public class BundleExtensionProvider {

    private final static String EXTENSIONS = "lib" + File.separator + "extensions";

    private final static String EXTENSION_PREFIX = "org.wso2.carbon.framework.extension.";

    private final static String VERSION = "1.0.0";

    public static void provideExtensions(File plugins) {
        try {
            String carbonHome = System.getProperty("carbon.home");

            if (carbonHome == null) {
                carbonHome = System.getenv("CARBON_HOME");
            }

            if (carbonHome == null || carbonHome.length() == 0) {
                return;
            }
            File carbonHomeFile = new File(carbonHome, EXTENSIONS);
            if (!carbonHomeFile.exists()) {
                carbonHomeFile.mkdirs();
            }
            File[] files = carbonHomeFile.listFiles(new JarFileFilter());
            if (files != null) {
                for (File file : files) {
                    //create the extension
                    String exportedPackages = parseJar(file);
                    createExtensionBundle(exportedPackages, file, plugins);
                }
            }
        } catch (IOException e) {
            //TODO debuginfo ?
            e.printStackTrace();
        }
    }

    private static class JarFileFilter implements FileFilter {
        public boolean accept(File pathname) {
            return pathname.getName().endsWith(".jar");
        }
    }

    /**
     * @param jarFile jar file location
     * @return package name list separated by ","
     * @throws IOException IOException
     */
    private static String parseJar(File jarFile) throws IOException {
        List<String> exportedPackagesList = new ArrayList<String>();
        ZipInputStream zipInputStream = new ZipInputStream(
                new BufferedInputStream(new FileInputStream(jarFile)));
        List<ZipEntry> entries = populateList(zipInputStream);
        zipInputStream.close();

        for (ZipEntry entry : entries) {
            String path = entry.getName();
            if (!path.endsWith("/") && path.endsWith(".class")) {
                //This is package that contains classes. Thus, exportedPackagesList
                int index = path.lastIndexOf("/");
                path = path.substring(0, index);
                path = path.replaceAll("/", ".");
                if (!exportedPackagesList.contains(path)) {
                    exportedPackagesList.add(path);
                }
            }
        }

        String[] packageArray =
                exportedPackagesList.toArray(new String[exportedPackagesList.size()]);
        StringBuffer exportedPackages = new StringBuffer();
        for (int i = 0; i < packageArray.length; i++) {
            exportedPackages.append(packageArray[i]);
            if (i != (packageArray.length -1)) {
                exportedPackages.append(",");
            }
        }
        return exportedPackages.toString();
    }

    /**
     * @param zipInputStream zipInputStream
     * @return return zipetry map
     * @throws IOException IOException
     */
    private static List<ZipEntry> populateList(ZipInputStream zipInputStream) throws IOException {
        List<ZipEntry> listEntry = new ArrayList<ZipEntry>();
        while (zipInputStream.available() == 1) {
            ZipEntry entry = zipInputStream.getNextEntry();
            if (entry == null) {
                break;
            }
            listEntry.add(entry);
        }
        return listEntry;
    }

    private static void createExtensionBundle(String exportedPackageNames, File jarFile,
                                              File plugins)
            throws IOException {
        String symbolicName = EXTENSION_PREFIX + System.currentTimeMillis() + Math.random();
        String fileName = jarFile.getName();
        fileName = fileName.replaceAll("-", "_");
        String pluginName = EXTENSION_PREFIX + fileName;
        File extensionBundle = new File(plugins, pluginName);
        if (extensionBundle.exists()) {
            return;
        }
        Manifest mf = new Manifest();
        Attributes attribs = mf.getMainAttributes();
        attribs.putValue(MANIFEST_VERSION, "1.0");
        attribs.putValue(BUNDLE_MANIFEST_VERSION, "2");
        attribs.putValue(BUNDLE_NAME, "Extension Bundle for " + fileName);
        attribs.putValue(BUNDLE_SYMBOLIC_NAME, symbolicName);
        attribs.putValue(BUNDLE_VERSION, VERSION);
        attribs.putValue(FRAGMENT_HOST, "system.bundle; extension:=framework");

        attribs.putValue(EXPORT_PACKAGE, exportedPackageNames);

        try {
            JarOutputStream jos = null;
            try {
                jos = new JarOutputStream(new FileOutputStream(extensionBundle), mf);
                jos.finish();
            } finally {
                if (jos != null)
                    jos.close();
            }
            //install the master bundle
        } catch (IOException e) {
            throw e;
        }
    }
}
