/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.core.services.authentication;

import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpSession;

import org.apache.axis2.context.MessageContext;
import org.apache.axis2.transport.http.HTTPConstants;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.user.core.Authorizer;
import org.wso2.carbon.user.core.SystemPermissions;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.utils.ServerConstants;
import org.wso2.carbon.registry.core.service.RegistryService;
import org.wso2.carbon.registry.core.session.UserRegistry;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.sso.consumer.SSOConstants;
import org.wso2.carbon.sso.consumer.SSOConsumer;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;


/**
 * Admin service to manage operations on the <code>AxisConfiguration</code>
 */
public class AuthenticationAdmin implements AuthenticationInterface {

    private static final Log log = LogFactory.getLog(AuthenticationAdmin.class);

    public boolean login(String username, String password, String remoteAddress) throws AuthenticationException {

        MessageContext msgCtx = MessageContext.getCurrentMessageContext();
        HttpServletRequest request =
                (HttpServletRequest) msgCtx.getProperty(HTTPConstants.MC_HTTP_SERVLETREQUEST);
        HttpSession httpSess = request.getSession();

        Date currentTime = Calendar.getInstance().getTime();
        SimpleDateFormat date = new SimpleDateFormat("'['yyyy-MM-dd HH:mm:ss,SSSS']'");

        if ((username == null) || (password == null) || (remoteAddress == null) ||
                username.trim().equals("") || password.trim().equals("") ||
                remoteAddress.trim().equals("")) {
            log.warn("Failed Administrator login attempt \'"
                    + username + "\' at " + date.format(currentTime) + " from IP address "
                    + remoteAddress);
            httpSess.invalidate();
            return false;
        }

        try {

			AuthenticationAdminTrackers trackers = new AuthenticationAdminTrackers();
			UserRealm realm = trackers.getUserRealm();
			boolean isAuthenticated = false;

			if (MessageContext.getCurrentMessageContext().getAxisService().getAxisConfiguration()
					.getParameter(SSOConstants.SSO_SERVICE_EPR) != null) {
				String epr = (String) MessageContext.getCurrentMessageContext().getAxisService()
						.getAxisConfiguration().getParameter(SSOConstants.SSO_SERVICE_EPR)
						.getValue();
				// If this is the case - what we get for password is considered as the session id
				// for SSO.
				isAuthenticated = new SSOConsumer(epr).isAuthenticated(username, password);
			}
			
			if (!isAuthenticated) {
				isAuthenticated = realm.getAuthenticator().authenticate(username, password);
			} else {
				if (log.isDebugEnabled()) {
					log.debug("Logged-in with SSO enabled login");
				}
			}
			
			boolean isAuthorized = realm.getAuthorizer().isUserAuthorized(username, "System",
					"login");
			if (isAuthenticated && isAuthorized) {
				httpSess.setAttribute(ServerConstants.ADMIN_LOGGED_IN, "true");
				log.info("Administrator logged in \'" + username + "\' at " + date.format(currentTime) + " from IP address "
                        + remoteAddress);

                // create a UserRegistry instance for the logged in user. This instance is used to perform Registry
                // actions performed by *this* user. If a Registry action is performed by the system, it is
                // done using the system registry instance, althouth user's registry is available.

                // in most cases, user registry instance is used only by the Registry UI bundles as those actions
                // are performed directly by the currently logged in user.

                RegistryService registryService = trackers.getRegistryService();
                UserRegistry userRegistry = registryService.getUserRegistry(username);

                request.getSession().setAttribute(RegistryConstants.USER_REGISTRY, userRegistry);

                if (log.isDebugEnabled()) {
                    log.debug("User Registry instance is set in the session for user " + username);
                }

                return true;
            } else {
                httpSess.invalidate();
                log.warn("Failed Administrator login attempt \'"
                        + username + "\' at " + date.format(currentTime) + " from IP address "
                        + remoteAddress);
                return false;
            }
        } catch (Exception e) {
            String msg = "System error while Authenticating/Authorizing User";
            log.error(msg, e);
            return false;
        }
    }

    public void logout() throws AuthenticationException {
        log.info("Administrator logged out");
        MessageContext msgCxt = MessageContext.getCurrentMessageContext();
        HttpServletRequest request =
                (HttpServletRequest) msgCxt.getProperty(HTTPConstants.MC_HTTP_SERVLETREQUEST);
        HttpSession session = request.getSession();
        session.invalidate();
    }

    public String[] getSystemPermissionOfUser(String username) throws Exception {
        try {
            MessageContext msgCtx = MessageContext.getCurrentMessageContext();
            HttpServletRequest request =
                    (HttpServletRequest) msgCtx.getProperty(HTTPConstants.MC_HTTP_SERVLETREQUEST);
            HttpSession httpSess = request.getSession();
            AuthenticationAdminTrackers trackers = new AuthenticationAdminTrackers();
            UserRealm realm = trackers.getUserRealm();
            List<String> userPermissions = getUserPermissions(username, realm);
            String[] permissions = userPermissions.toArray(new String[userPermissions.size()]);
            return permissions;
        } catch(Exception e) {
            log.error(e);
            throw e;
        }
    }


    public List<String> getUserPermissions(String username, UserRealm realm) throws Exception{
        List<String> permissions = SystemPermissions.getSystemPermission();
        List<String> userPermissions = new ArrayList<String>();
        Authorizer authorizer = realm.getAuthorizer();
        for(String permission : permissions){
            if(authorizer.isUserAuthorized(username, "System", permission)){
                userPermissions.add(permission);
            }
        }
        return userPermissions;
    }
}
