/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.internal;

import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;
import org.osgi.util.tracker.ServiceTracker;
import org.wso2.carbon.registry.core.jdbc.EmbeddedRegistry;
import org.wso2.carbon.registry.core.jdbc.realm.RegistryRealm;
import org.wso2.carbon.registry.core.jdbc.realm.CarbonRealm;
import org.wso2.carbon.registry.core.config.RegistryContext;
import org.wso2.carbon.registry.core.config.RegistryConfiguration;
import org.wso2.carbon.registry.core.exceptions.RegistryException;
import org.wso2.carbon.registry.core.service.RegistryService;
import org.wso2.carbon.registry.core.session.UserRegistry;
import org.wso2.carbon.registry.core.ResourcePath;
import org.wso2.carbon.registry.core.RegistryConstants;
import org.wso2.carbon.registry.core.Collection;
import org.wso2.carbon.registry.core.utils.AuthorizationUtils;
import org.wso2.carbon.registry.app.RemoteRegistryService;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.utils.ServerConfiguration;
import org.wso2.carbon.utils.CarbonUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.io.InputStream;
import java.io.FileInputStream;
import java.io.File;

public class Activator implements BundleActivator {

    private static final String CARBON_COLLECTION_NAME = "carbon";
    private static final String CARBON_COLLECTION_PARENT_PROPERTY = "carbon.registry.root";

    private static final Log log = LogFactory.getLog(Activator.class);

    public void start(BundleContext bundleContext) throws Exception {

        ServiceReference[] serviceRefs = bundleContext.
                getServiceReferences(UserRealm.class.getName(), "(RealmGenre=Default)");
        ServiceTracker realmTracker = new ServiceTracker(bundleContext, serviceRefs[0], null);
        realmTracker.open();
        CarbonRealm carbonRealm = new CarbonRealm(realmTracker, bundleContext);
        UserRealm registryRealm = new RegistryRealm(carbonRealm);

        ServerConfiguration serverConfig =  getServerConfiguration();
        setSystemTruststore(serverConfig);
        
        RegistryConfiguration regConfig = getRegistryConfiguration();
        RegistryService registryService;

        if (RegistryConfiguration.REMOTE_REGISTRY.equals(regConfig.getRegistryType())) {
            registryService = getRemoteRegistryService(registryRealm,regConfig);   
        } else {
            registryService = getEmbbedRegistryService(registryRealm);
        }
        log.info("Using " +  regConfig.getRegistryType() + " Registry");

        if (CarbonUtils.isRegistryReadOnly()) {
            log.info("Registry is used in READ-ONLY mode");
        } else {
            log.info("Registry is used in READ-WRITE mode");
        }

        bundleContext.registerService(RegistryService.class.getName(), registryService , null);
    }

    public void stop(BundleContext bundleContext) throws Exception {

    }

    private File getConfigFile() throws RegistryException {

        String carbonHome = System.getProperty("carbon.home");

        if (carbonHome == null) {
            carbonHome = System.getenv("CARBON_HOME");
        }

        if (carbonHome != null) {

            String configPath = carbonHome + File.separator +
                    "conf" + File.separator + "registry.xml";
            File registryXML = new File(configPath);
            if (!registryXML.exists()) {
                String msg = "Registry configuration file (registry.xml) file does " +
                        "not exist in the path " + configPath;
                log.error(msg);
                throw new RegistryException(msg);
            }

            return registryXML;

        } else {
            String msg = "carbon.home system property is not set. It is required to to derive " +
                    "the path of the Registry configuration file (registry.xml).";
            log.error(msg);
            throw new RegistryException(msg);
        }
    }

    private void addCarbonRootCollection(UserRegistry userRegistry) throws RegistryException {

        String carbonPathParent = System.getProperty(CARBON_COLLECTION_PARENT_PROPERTY);

        String carbonPath;
        if (carbonPathParent != null) {
            ResourcePath rootPath = new ResourcePath(carbonPathParent);
            rootPath.appendPath(CARBON_COLLECTION_NAME);
            carbonPath = rootPath.getPath();
        }

        carbonPath = RegistryConstants.ROOT_PATH + CARBON_COLLECTION_NAME;

        String carbonRoot;
        int index = carbonPath.indexOf(RegistryConstants.PATH_SEPARATOR, 1);
        if (index == -1) {
            carbonRoot = carbonPath;
        } else {
            carbonRoot = carbonPath.substring(0, index);
        }

        Collection carbonRootCollection = userRegistry.newCollection();
        String carbonRootDescription = "Root of the system collections used by the Carbon server.";
        carbonRootCollection.setDescription(carbonRootDescription);
        userRegistry.put(carbonRoot, carbonRootCollection);
        carbonRootCollection.discard();

        UserRealm userRealm = userRegistry.getUserRealm();
        AuthorizationUtils.setAdminAuthorizations(carbonRoot, userRealm);

        Collection carbonCollection = userRegistry.newCollection();
        userRegistry.put(carbonPath, carbonCollection);
        carbonCollection.discard();
    }

    private RegistryService getEmbbedRegistryService(UserRealm registryRealm) throws Exception {

        InputStream configInputStream = new FileInputStream(getConfigFile());
        RegistryContext registryContext = new RegistryContext(configInputStream, registryRealm);
        registryContext.setSetup(System.getProperty(RegistryConstants.SETUP_PROPERTY) != null);
        EmbeddedRegistry embeddedRegistry = new EmbeddedRegistry(registryContext);

        UserRegistry systemRegistry =  embeddedRegistry.getSystemRegistry();
        addCarbonRootCollection(systemRegistry);

        return embeddedRegistry;
    }

    private RegistryService getRemoteRegistryService(UserRealm registryRealm, RegistryConfiguration regConfig) throws Exception {

        String url = regConfig.getValue(RegistryConfiguration.URL);
        String username = regConfig.getValue(RegistryConfiguration.USERNAME);
        String password = regConfig.getValue(RegistryConfiguration.PASSWORD);
        String chroot = regConfig.getValue(RegistryConfiguration.CHROOT); 

        RemoteRegistryService registryService = new RemoteRegistryService(url,username,password,registryRealm,chroot);

        UserRegistry systemRegistry =  registryService.getSystemRegistry();
        addCarbonRootCollection(systemRegistry);

        return registryService;
    }

    private RegistryConfiguration getRegistryConfiguration() throws RegistryException {

        String path = getCarbonXMLPath();
        return new RegistryConfiguration(path);

    }

    private String getCarbonXMLPath() {
        String carbonXML = System.getProperty("wso2.carbon.xml");

        if (carbonXML == null) {
            String carbonHome = System.getProperty("carbon.home");
            if (carbonHome == null) {
                carbonHome = System.getenv("CARBON_HOME");
            }

            if (carbonHome != null) {
                carbonXML = carbonHome + File.separator + "conf" +
                        File.separator + "carbon.xml";
            }
        }
        return carbonXML;
    }

    private void setSystemTruststore(ServerConfiguration serverConfiguration) throws RegistryException {

        if (serverConfiguration != null) {
            String location = serverConfiguration.getFirstProperty("Security.KeyStore.Location");
            String password = serverConfiguration.getFirstProperty("Security.KeyStore.Password");
            String type = serverConfiguration.getFirstProperty("Security.KeyStore.Type");

            if (location != null && password != null) {
                System.setProperty("javax.net.ssl.trustStore", location);
                System.setProperty("javax.net.ssl.trustStoreType", type);
                System.setProperty("javax.net.ssl.trustStorePassword", password);
            }

        }

    }

    private ServerConfiguration getServerConfiguration() throws RegistryException {
        ServerConfiguration serverConfig = ServerConfiguration.getInstance();
        try {
            File carbonXML = new File(CarbonUtils.getServerXml());
            InputStream inSXml = new FileInputStream(carbonXML);
            serverConfig.init(inSXml);
        } catch (Exception e) {
            log.error("Error initializing the Server Configuration", e);
            throw new RegistryException("Error initializing the Server Configuration", e);
        }

        return serverConfig;
    }

}
