/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.registry.core.jdbc.utils.creators;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.registry.core.exceptions.RegistryException;

import javax.sql.DataSource;
import java.sql.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

public abstract class DatabaseCreator {

    private static Log log = LogFactory.getLog(DatabaseCreator.class);

    private List <String> resourceTableNames = new ArrayList <String> ();
    private List <String> userTableNames = new ArrayList <String> ();

    private DataSource dataSource;

    protected DatabaseCreator(DataSource dataSource) {
        this.dataSource = dataSource;

        resourceTableNames.add("REG_RESOURCE");
        resourceTableNames.add("REG_CONTENT");
        resourceTableNames.add("REG_DEPENDENCY");
        resourceTableNames.add("REG_DEPENDENCY");
        resourceTableNames.add("REG_PROPERTY");
        resourceTableNames.add("REG_ASSOCIATION");
        resourceTableNames.add("REG_TAG");
        resourceTableNames.add("REG_COMMENT");
        resourceTableNames.add("REG_RATING");
        resourceTableNames.add("REG_LOG");
        resourceTableNames.add("REG_RESOURCE_VERSION");
        resourceTableNames.add("REG_DEPENDENCY_VERSION");
        resourceTableNames.add("REG_CONTENT_VERSION");
        resourceTableNames.add("REG_PROPERTY_VERSION");
        resourceTableNames.add("REG_SNAPSHOT");
        resourceTableNames.add("REG_SNAPSHOT_RESOURCE_VERSION");

        userTableNames.add("UM_USERS");
        userTableNames.add("UM_USER_ATTRIBUTES");
        userTableNames.add("UM_ROLES");
        userTableNames.add("UM_ROLE_ATTRIBUTES");
        userTableNames.add("UM_PERMISSIONS");
        userTableNames.add("UM_USER_PERMISSIONS");
        userTableNames.add("UM_ROLE_PERMISSIONS");
    }

    /**
     * Creates the database tables required for registry (resource store), if they are not already
     * created
     */
    public void createRegistryDatabase() throws RegistryException {

        Connection conn = null;

        try {

            conn = dataSource.getConnection();
            conn.setAutoCommit(false);

            DatabaseMetaData metaData = conn.getMetaData();
            ResultSet result = metaData.getTables(null, null, "REG_RESOURCE", null);
            if (result.next()) {
                if (log.isDebugEnabled()) {
                    log.debug("Registry tables were already created in the database.");
                }
                return;
            }

            ResultSet result2 = metaData.getTables(null, null, "reg_resource", null);
            if (result2.next()) {
                if (log.isDebugEnabled()) {
                    log.debug("Registry tables were already created in the database.");
                }
                return;
            }

            Statement s = conn.createStatement();

            // create current version related tables
            s.executeUpdate(getContentTable());
            s.executeUpdate(getResourceTable());
            s.executeUpdate(getDependencyTable());
            s.executeUpdate(getPropertyTable());
            s.executeUpdate(getAssociationTable());
            s.executeUpdate(getTagTable());
            s.executeUpdate(getCommentTable());
            s.executeUpdate(getRatingTable());
            s.executeUpdate(getLogTable());

            // create old versions related tables
            s.executeUpdate(getContentVersionTable());
            s.executeUpdate(getResourceVersionTable());
            s.executeUpdate(getDependencyVersionTable());
            s.executeUpdate(getPropertyVersionTable());

            // create snapshot management tables
            s.executeUpdate(getSnapshotTable());
            s.executeUpdate(getSnapshotRootIDIndex());
            s.executeUpdate(getSnapshotResourceVersionTable());

            executeResourceStoreDatabaseSpecificStatements(s);

            conn.commit();
            if (log.isTraceEnabled()) {
                log.trace("Registry tables are created successfully.");
            }

        } catch (SQLException e) {

            // Note if Registry is running on multiple nodes in a cluster sharing a same database:
            // When nodes are started at the same time, the first node that reaches database
            // creation code, tries to create tables. But other nodes may also perform the check and
            // finds out that tables are not created (if the first node has not yet commited the
            // table creation operation). Then they may also try to create tables and only one node
            // succeeds in this process. All other nodes may fail in this table creation operation,
            // if they encounter this concurrency scenario. But all other nodes should continue,
            // if the tables are created by some other node.

            boolean resourceTablesCreated = false;
            try {
                if (conn != null && isResourceTablesCreated(conn)) {
                    resourceTablesCreated = true;
                }
            } catch (SQLException e1) {
                String msg = "Failed to check the existance of resource tables. Caused by: " +
                        e1.getMessage();
                log.error(msg, e1);
            }

            if (resourceTablesCreated) {
                String msg = "Attempt to create database tables for resource store is " +
                        "unsuccessful. But all the required database tables are already created. " +
                        "If this Registry instance is started in a clustered environment " +
                        "simultaneously with other Registry nodes, this is an expected situation " +
                        "and considered as a successful startup of this node. Complete error " +
                        "is also logged for reference. Ignore this error log if this " +
                        "instance is started in a cluster.";
                log.error(msg, e);

            } else {

                String msg = "Failed to create database tables for registry resource store. " +
                        e.getMessage();
                log.fatal(msg, e);

                try {
                    if (conn != null) {
                        conn.rollback();
                    }
                } catch (SQLException e1) {
                    log.error("Failed to rollback the transaction used for creating registry " +
                            "database tables.");
                }

                throw new RegistryException(msg, e);
            }

        } finally {
            try {
                if (conn != null) {
                    conn.close();
                }
            } catch (SQLException e) {
                log.error("Failed to close the database connection.");
            }
        }

    }

    /**
     * Creates the database tables required for user manager (user store), is they are not already
     * created.
     */
    public void createUserManagerDatabase() throws RegistryException {

        Connection conn = null;

        try {

            conn = dataSource.getConnection();
            conn.setAutoCommit(false);

            DatabaseMetaData metaData = conn.getMetaData();
            ResultSet result = metaData.getTables(null, null, "UM_USERS", null);
            if (result.next()) {
                if (log.isDebugEnabled()) {
                    log.debug("User Manager tables were already created in the database.");
                }
                return;
            }

            ResultSet result2 = metaData.getTables(null, null, "um_users", null);
            if (result2.next()) {
                if(log.isDebugEnabled()) {
                    log.debug("User Manager tables were already created in the database.");
                }
                return;
            }

            Statement s = conn.createStatement();

            s.executeUpdate(getUserTable());
            s.executeUpdate(getRoleTable());
            s.executeUpdate(getUserAttributeTable());
            s.executeUpdate(getRoleAttributeTable());
            s.executeUpdate(getPermissionTable());
            s.executeUpdate(getRolePermissionTable());
            s.executeUpdate(getUserPermissionTable());
            s.executeUpdate(getUserRolesTable());

            executeUserStoreDatabaseSpecificStatements(s);

            conn.commit();

            if (log.isTraceEnabled()) {
                log.trace("User Manager tables are created successfully.");
            }


        } catch (SQLException e) {

            // Note if Registry is running on multiple nodes in a cluster sharing a same database:
            // When nodes are started at the same time, the first node that reaches database
            // creation code, tries to create tables. But other nodes may also perform the check and
            // finds out that tables are not created (if the first node has not yet commited the
            // table creation operation). Then they may also try to create tables and only one node
            // succeeds in this process. All other nodes may fail in this table creation operation,
            // if they encounter this concurrency scenario. But all other nodes should continue,
            // if the tables are created by some other node.

            boolean userTablesCreated = false;
            try {
                if (conn != null && isUserTablesCreated(conn)) {
                    userTablesCreated = true;
                }
            } catch (SQLException e1) {
                String msg = "Failed to check the existance of user tables. Caused by: " +
                        e1.getMessage();
                log.error(msg, e1);
            }

            if (userTablesCreated) {
                String msg = "Attempt to create database tables for user store is " +
                        "unsuccessful. But all the required database tables are already created. " +
                        "If this Registry instance is started in a clustered environment " +
                        "simultaneously with other Registry nodes, this is an expected situation " +
                        "and considered as a successful startup of this node. Complete error " +
                        "is also logged for reference. Ignore this error log if this " +
                        "instance is started in a cluster.";
                log.error(msg, e);

            } else {

                String msg = "Failed to create database tables for user manager. " +
                        e.getMessage();
                log.fatal(msg, e);

                try {
                    if (conn != null) {
                        conn.rollback();
                    }
                } catch (SQLException e1) {
                    log.error("Failed to rollback the transaction used for " +
                            "creating user manager database tables.");
                }

                throw new RegistryException(msg, e);
            }

        } finally {
            try {
                if (conn != null) {
                    conn.close();
                }
            } catch (SQLException e) {
                log.error("Failed to close the database connection.");
            }
        }
    }

    public abstract String getClusterLockTable();

    public abstract String getResourceTable();

    public abstract String getContentTable();

    public abstract String getDependencyTable();

    public abstract String getPropertyTable();

    public abstract String getAssociationTable();

    public abstract String getTagTable();

    public abstract String getCommentTable();

    public abstract String getRatingTable();

    public abstract String getLogTable();

    public abstract String getResourceVersionTable();

    public abstract String getDependencyVersionTable();

    public abstract String getContentVersionTable();

    public abstract String getPropertyVersionTable();

    public abstract String getSnapshotTable();

    public abstract String getSnapshotResourceVersionTable();

    public abstract String getSnapshotRootIDIndex();

    public abstract String getUserTable();

    public abstract String getUserAttributeTable();

    public abstract String getRoleTable();

    public abstract String getRoleAttributeTable();

    public abstract String getPermissionTable();

    public abstract String getRolePermissionTable();

    public abstract String getUserPermissionTable();

    public abstract String getUserRolesTable();

    public abstract void executeResourceStoreDatabaseSpecificStatements(Statement s)
            throws SQLException;

    public abstract void executeUserStoreDatabaseSpecificStatements(Statement s)
            throws SQLException;

    private boolean isResourceTablesCreated(Connection conn) throws SQLException {

        boolean resourceTablesExist = true;
        Iterator <String> tables = resourceTableNames.iterator();
        while (tables.hasNext()) {
            String tableName = tables.next();

            if (!isTableExists(conn, tableName)) {
                if (log.isTraceEnabled()) {
                    log.trace(tableName + " table does not exist in the Registry database.");
                }
                resourceTablesExist = false;
                break;
            }
        }

        return resourceTablesExist;
    }

    private boolean isUserTablesCreated(Connection conn) throws SQLException {

        boolean userTablesExist = true;
        Iterator <String> tables = userTableNames.iterator();
        while (tables.hasNext()) {
            String tableName = tables.next();

            if (!isTableExists(conn, tableName)) {
                if (log.isTraceEnabled()) {
                    log.trace(tableName + " table does not exist in the Registry database.");
                }
                userTablesExist = false;
                break;
            }
        }

        return userTablesExist;
    }

    private boolean isTableExists(Connection conn, String tabelName) throws SQLException {

        DatabaseMetaData metaData = conn.getMetaData();

        boolean tableExists = false;
        ResultSet result = metaData.getTables(null, null, tabelName, null);
        if (result.next()) {
            if (log.isTraceEnabled()) {
                log.trace(tabelName + " table exists in the Registry database.");
            }
            tableExists = true;
        }
        result.close();

        return tableExists;
    }
}
