/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.user.core.def;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.carbon.user.core.UserStoreException;
import org.wso2.carbon.user.core.UserStoreReader;
import org.wso2.carbon.user.core.def.util.DefaultCommons;
import org.wso2.carbon.user.core.def.util.DefaultSQLHelperImpl;
import org.wso2.carbon.user.core.i18n.Messages;

/**
 * Manipulates data in the User Store
 * 
 * @see org.wso2.usermanager.UserStoreReader
 */
public class DefaultUserStoreReader implements UserStoreReader {

    private static Log log = LogFactory.getLog(DefaultUserStoreReader.class);

    protected DataSource dataSource = null;

    protected SQLHelper sqlHelper = null;

    public DefaultUserStoreReader(DataSource dataSource) {
	this.dataSource = dataSource;
	this.sqlHelper = new DefaultSQLHelperImpl();
    }

    public DefaultUserStoreReader(DataSource dataSource, SQLHelper sqlHelper) {
	this.dataSource = dataSource;
	this.sqlHelper = sqlHelper;
    }

    public String[] getAllRoleNames() throws UserStoreException {
	String[] names = new String[0];
	Connection dbConnection = null;
	try {
	    dbConnection = dataSource.getConnection();
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    PreparedStatement getAllRoleNamesStmt = dbConnection.prepareStatement(sqlHelper
		    .getSQL(DefaultRealmConstants.GET_ROLES_ALL));
	    ResultSet rs = getAllRoleNamesStmt.executeQuery();
	    List lst = new LinkedList();
	    String colName = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_ROLE_NAME);
	    while (rs.next()) {
		lst.add(rs.getString(colName));
	    }
	    if (lst.size() > 0) {
		names = (String[]) lst.toArray(new String[lst.size()]);
	    }
	    getAllRoleNamesStmt.close();
	} catch (SQLException e) {
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}
	return names;
    }

    public String[] getAllUserNames() throws UserStoreException {
	String[] names = new String[0];
	Connection dbConnection = null;
	try {
	    dbConnection = dataSource.getConnection();
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    PreparedStatement getAllUserNamesStmt = dbConnection.prepareStatement(sqlHelper
		    .getSQL(DefaultRealmConstants.GET_USERS_ALL));
	    ResultSet rs = getAllUserNamesStmt.executeQuery();
	    List lst = new LinkedList();
	    String colName = DefaultRealmConstants.COLUMN_NAME_USER_NAME;
	    while (rs.next()) {
		String name = rs.getString(colName);
		lst.add(name);
	    }
	    if (lst.size() > 0) {
		names = (String[]) lst.toArray(new String[lst.size()]);
	    }
	    getAllUserNamesStmt.close();
	} catch (SQLException e) {
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}
	return names;
    }

    public Map getRoleProperties(String roleName) throws UserStoreException {

	Map props = new HashMap();
	Connection dbConnection = null;
	try {
	    dbConnection = dataSource.getConnection();
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    PreparedStatement getRolePropertiesStmt = dbConnection.prepareStatement(sqlHelper
		    .getSQL(DefaultRealmConstants.GET_ROLE_ATTRIBUTES));
	    getRolePropertiesStmt.setString(1, roleName);
	    ResultSet rs = getRolePropertiesStmt.executeQuery();
	    String colAttrName = DefaultRealmConstants.COLUMN_NAME_ATTR_NAME;
	    String colAttrValue = DefaultRealmConstants.COLUMN_NAME_ATTR_VALUE;
	    while (rs.next()) {
		String name = rs.getString(colAttrName);
		String value = rs.getString(colAttrValue);
		props.put(name, value);
	    }
	    getRolePropertiesStmt.close();
	} catch (SQLException e) {
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}
	return props;

    }

    public Map getUserProperties(String userName) throws UserStoreException {
	Map props = new HashMap();
	Connection dbConnection = null;
	try {
	    dbConnection = dataSource.getConnection();
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    PreparedStatement getUserPropertiesStmt = dbConnection.prepareStatement(sqlHelper
		    .getSQL(DefaultRealmConstants.GET_USER_ATTRIBUTES));
	    getUserPropertiesStmt.setString(1, userName);
	    ResultSet rs = getUserPropertiesStmt.executeQuery();
	    String colAttrName = DefaultRealmConstants.COLUMN_NAME_ATTR_NAME;
	    String colAttrValue = DefaultRealmConstants.COLUMN_NAME_ATTR_VALUE;
	    while (rs.next()) {
		String name = rs.getString(colAttrName);
		String value = rs.getString(colAttrValue);
		props.put(name, value);
	    }
	    getUserPropertiesStmt.close();
	} catch (SQLException e) {
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}
	return props;

    }

    public String[] getUserPropertyNames() throws UserStoreException {
	String[] propNames = new String[0];
	Connection dbConnection = null;
	try {
	    dbConnection = dataSource.getConnection();
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    PreparedStatement getUserAttributeNamesStmt = dbConnection.prepareStatement(sqlHelper
		    .getSQL(DefaultRealmConstants.GET_ATTRIBUTE_NAMES));
	    ResultSet rs = getUserAttributeNamesStmt.executeQuery();
	    String colName = DefaultRealmConstants.COLUMN_NAME_ATTR_NAME;
	    List lst = new ArrayList();
	    while (rs.next()) {
		lst.add(rs.getString(colName));
	    }
	    propNames = (String[]) lst.toArray(new String[lst.size()]);
	    getUserAttributeNamesStmt.close();
	} catch (SQLException e) {
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}
	return propNames;
    }

    public String[] getUserRoles(String userName) throws UserStoreException {
	String[] names = new String[0];
	Connection dbConnection = null;
	try {
	    dbConnection = dataSource.getConnection();
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    DefaultCommons commons = new DefaultCommons();
	    names = commons.getUserRoles(sqlHelper, dbConnection, userName);
	} catch (SQLException e) {
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}
	return names;
    }

    public boolean isExistingUser(String userName) throws UserStoreException {
	boolean isExisting = false;
	Connection dbConnection = null;
	try {
	    dbConnection = dataSource.getConnection();
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    DefaultCommons commons = new DefaultCommons();
	    if (commons.getUserId(sqlHelper, dbConnection, userName) != -1) {
		isExisting = true;
	    }
	} catch (SQLException e) {
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}
	return isExisting;
    }

    public boolean isExistingRole(String roleName) throws UserStoreException {
	boolean isExisting = false;
	Connection dbConnection = null;
	try {
	    dbConnection = dataSource.getConnection();
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    DefaultCommons commons = new DefaultCommons();
	    if (commons.getRoleId(sqlHelper, dbConnection, roleName) != -1) {
		isExisting = true;
	    }
	} catch (SQLException e) {
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}

	return isExisting;
    }

    /**
     * If propertyName is null then the value is searched If propertyName
     * contains % then a likenss querry is searched
     * 
     */
    public String[] getUserNamesWithPropertyValue(String propertyName, String propetyValue)
	    throws UserStoreException {
	String[] names = new String[0];
	Connection dbConnection = null;
	try {
	    dbConnection = dataSource.getConnection();
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    PreparedStatement getPropValue = null;

	    if (propertyName != null && propertyName.contains("%")) {
		getPropValue = dbConnection.prepareStatement(sqlHelper
			.getSQL(DefaultRealmConstants.GET_USERS_WITH_PROPERTY_LIKE));
		getPropValue.setString(1, propertyName);
		getPropValue.setString(2, propetyValue);
	    } else if (propertyName != null && propertyName.length() != 0) {
		getPropValue = dbConnection.prepareStatement(sqlHelper
			.getSQL(DefaultRealmConstants.GET_USERS_WITH_PROPERTY));
		getPropValue.setString(1, propertyName);
		getPropValue.setString(2, propetyValue);
	    } else {
		getPropValue = dbConnection.prepareStatement(sqlHelper
			.getSQL(DefaultRealmConstants.GET_USERS_WITH_PROPERTY_VALUE));
		getPropValue.setString(1, propetyValue);
	    }

	    ResultSet rs = getPropValue.executeQuery();
	    String colName = DefaultRealmConstants.COLUMN_NAME_USER_NAME;
	    List lst = new ArrayList();
	    while (rs.next()) {
		lst.add(rs.getString(colName));
	    }
	    names = (String[]) lst.toArray(new String[lst.size()]);
	    getPropValue.close();
	} catch (SQLException e) {
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}
	return names;

    }

    public String[] getUsersInRole(String roleName) throws UserStoreException {
	String[] names = new String[0];
	Connection dbConnection = null;
	try {
	    dbConnection = dataSource.getConnection();
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    PreparedStatement getUserInRole = dbConnection.prepareStatement(sqlHelper
		    .getSQL(DefaultRealmConstants.GET_USERS_IN_ROLE));
	    getUserInRole.setString(1, roleName);
	    ResultSet rs = getUserInRole.executeQuery();
	    String colName = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_USER_NAME);
	    List lst = new ArrayList();
	    while (rs.next()) {
		lst.add(rs.getString(colName));
	    }
	    names = (String[]) lst.toArray(new String[lst.size()]);
	    getUserInRole.close();
	} catch (SQLException e) {
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}
	return names;
    }

    public Map<String, String> getUserProperties(String userName, String profileName)
	    throws UserStoreException {
	// TODO 1.0
	return null;
    }

    public List<String> getUserProfileNames(String userName) throws UserStoreException {
	// TODO 1.0
	return null;
    }

    public String getDefaultUserProfileName(String userName) throws UserStoreException {
	// TODO 1.0
	return null;
    }

    public boolean isExistingUserProfile(String userName, String profileName)
	    throws UserStoreException {
	//TODO 1.0
	return false;
    }

    public String[] listUsers(String filter, int maxItemLimit) throws UserStoreException {
	String[] users = new String[0];
	Connection dbConnection = null;
	try {
	    String sqlStmt = sqlHelper.getSQL(DefaultRealmConstants.GET_USERS_ALL);
	    if (filter != null && filter.trim().length() != 0) {
		filter = filter = filter.trim();
		filter = filter.replace("*", "%");
		filter = filter.replace("?", "_");
		sqlStmt = sqlStmt + " WHERE USER_NAME LIKE '" +filter+"' ORDER BY USER_NAME";
	    } else {
		sqlStmt = sqlStmt + " ORDER BY USER_NAME";
	    }

	    dbConnection = dataSource.getConnection();
	    if (dbConnection == null) {
		throw new UserStoreException(Messages.getMessage("null_connection"));
	    }
	    dbConnection.setAutoCommit(false);
	    dbConnection.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);

	    Statement stmt = dbConnection.createStatement();
	    ResultSet rs = stmt.executeQuery(sqlStmt);

	    List lst = new LinkedList();
	    String colName = sqlHelper.getColumnName(DefaultRealmConstants.COLUMN_ID_USER_NAME);

	    int i = 0;
	    while (rs.next()) {
		if (i < maxItemLimit) {
		    lst.add(rs.getString(colName));
		}else{
		    break;
		}
		i++;
	    }
	    rs.close();

	    if (lst.size() > 0) {
		users = (String[]) lst.toArray(new String[lst.size()]);
	    }

	} catch (SQLException e) {
	    log.debug(e);
	    throw new UserStoreException(Messages.getMessage("errorReadingFromUserStore"), e);
	} finally {
	    try {
		if (dbConnection != null) {
		    dbConnection.close();
		}
	    } catch (SQLException e) {
		throw new UserStoreException(Messages.getMessage("errorClosingConnection"), e);
	    }
	}
	return users;
    }

}
