/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.carbon.user.core.hybrid;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.Statement;
import java.util.Arrays;
import java.util.Properties;

import junit.framework.TestCase;

import org.apache.commons.dbcp.BasicDataSource;
import org.wso2.carbon.user.core.AccessControlAdmin;
import org.wso2.carbon.user.core.Authenticator;
import org.wso2.carbon.user.core.Authorizer;
import org.wso2.carbon.user.core.UserRealm;
import org.wso2.carbon.user.core.UserStoreAdmin;
import org.wso2.carbon.user.core.UserStoreReader;
import org.wso2.carbon.user.core.dbcreators.DerbyDatabaseCreator;
import org.wso2.carbon.user.core.hybrid.jdbc.JDBCAuthenticator;
import org.wso2.carbon.user.core.hybrid.jdbc.JDBCRealmConfig;
import org.wso2.carbon.user.core.hybrid.jdbc.JDBCUserStoreAdmin;
import org.wso2.carbon.user.core.hybrid.jdbc.JDBCUserStoreReader;

public class HybridRealmTest extends TestCase {

    UserRealm realm = null;

    HybridRealmConfig config = null;

    String connURL = "jdbc:derby:target/HybridUserDatabase";

    protected void setUp() throws Exception {
	super.setUp();
	// create the custom database
	createCustomDatabase();
	BasicDataSource ds = new BasicDataSource();
	ds.setUrl(connURL + ";create=true");
	ds.setUsername("roshan");
	ds.setUsername("jimmy");
	ds.setDriverClassName("org.apache.derby.jdbc.EmbeddedDriver");

	DerbyDatabaseCreator creator = new DerbyDatabaseCreator(ds);
	creator.createHybridDatabaseTables();

	JDBCRealmConfig jconfig = new JDBCRealmConfig();
	jconfig.setConnectionURL("jdbc:derby:target/ComapnyDataBase");
	jconfig.setConnectionPassword("snowy");
	jconfig.setConnectionName("tintin");
	jconfig.setRoleListSQL("select role_name from roles");
	jconfig.setUserListSQL("select employee_name from employees");
	jconfig
		.setUserFilterSQL("select employee_name from employees where employee_name like ?");
	jconfig
		.setUserRoleSQL("select role_name from employee_roles, employees, roles where employee_roles.role_id = roles.id AND employee_roles.emp_id = employees.id AND employees.employee_name=?");
	jconfig
		.setUsersInRoleSQL("select employee_name from employee_roles, employees, roles where employee_roles.role_id = roles.id AND employee_roles.emp_id = employees.id AND roles.role_name=?");
	
	Authenticator authenticator = new JDBCAuthenticator(jconfig);
	UserStoreReader reader = new JDBCUserStoreReader(jconfig);
	UserStoreReader admin = new JDBCUserStoreAdmin(jconfig);

	BasicDataSource dataSource = new BasicDataSource();
	dataSource.setUrl(connURL);
	dataSource.setUsername("roshan");
	dataSource.setUsername("jimmy");
	dataSource.setDriverClassName("org.apache.derby.jdbc.EmbeddedDriver");

	realm = new HybridRealm();
	config = (HybridRealmConfig) realm.getRealmConfiguration();
	config.setAuthenticator(authenticator);
	config.setDataSource(dataSource);
	config.setUserStoreReader(reader);

	realm.init(config);

    }

    public void testRealm() throws Exception {
	roleStuff();
	authorizationStuff();
	// roleUserStuff();
	internalRolesStuff();
    }

    public void roleStuff() throws Exception {

	UserStoreAdmin admin = realm.getUserStoreAdmin();
	UserStoreReader reader = realm.getUserStoreReader();
	// list
	String[] names = reader.getAllRoleNames();
	System.out.println(Arrays.toString(names));
	TestCase.assertEquals(2, names.length);

	String[] usernames = reader.getAllUserNames();
	TestCase.assertEquals(3, usernames.length);

	String[] usersInRole = reader.getUsersInRole("developer");

    }

    public void authorizationStuff() throws Exception {

	UserStoreAdmin admin = realm.getUserStoreAdmin();

	AccessControlAdmin acAdmin = realm.getAccessControlAdmin();
	Authorizer athzr = realm.getAuthorizer();
	Authenticator authen = realm.getAuthenticator();

	acAdmin.authorizeUser("dimuthu", "floor", "dance");
	acAdmin.authorizeRole("admin", "server", "login");
	acAdmin.authorizeUser("vajira", "floor", "dance");
	acAdmin.denyUser("dimuthu", "denyResource", "deny");
	acAdmin.authorizeRole("admin", "denyResource", "deny");
	acAdmin.denyRole("admin", "denyResource", "deny");

	TestCase.assertTrue(athzr.isUserAuthorized("dimuthu", "floor", "dance"));
	TestCase.assertTrue(athzr.isRoleAuthorized("admin", "server", "login"));
	TestCase.assertFalse(athzr.isUserAuthorized("dimuthu", "denyResource", "deny"));
	TestCase.assertFalse(athzr.isRoleAuthorized("admin", "denyResource", "deny"));

	String[] authorizedUsers = athzr.getAllowedUsersForResource("floor", "dance");
	String au = Arrays.toString(authorizedUsers);
	TestCase.assertTrue(au.contains("vajira"));
	TestCase.assertTrue(au.contains("dimuthu"));

	String[] authorizedRoles = athzr.getAllowedRolesForResource("server", "login");
	String ar = Arrays.toString(authorizedRoles);
	TestCase.assertTrue(ar.contains("admin"));

	String[] deniedRoles = athzr.getDeniedRolesForResource("denyResource", "deny");
	TestCase.assertEquals(deniedRoles[0], "admin");

	acAdmin.clearUserAuthorization("dimuthu", "floor", "dance");

	String[] deniedUsers = athzr.getDeniedUsersForResource("denyResource", "deny");
	TestCase.assertEquals(1, deniedUsers.length);

	TestCase.assertTrue(athzr.isRoleAuthorized("admin", "server", "login"));

	acAdmin.clearRoleAuthorization("admin", "server", "login");

	TestCase.assertFalse(athzr.isRoleAuthorized("admin", "server", "login"));

	TestCase.assertTrue(athzr.isUserAuthorized("vajira", "floor", "dance"));

	acAdmin.authorizeUser("juhia", "floor", "dance");
	TestCase.assertFalse(athzr.isUserAuthorized("admin", "server", "login"));
	acAdmin.authorizeRole("admin", "server", "login");
	System.out.println("Basic done");

	acAdmin.copyAuthorizations("floor", "classroom");
	TestCase.assertTrue(athzr.isUserAuthorized("juhia", "classroom", "dance"));

	acAdmin.copyAuthorizations("server", "desktop");
	TestCase.assertTrue(athzr.isRoleAuthorized("admin", "desktop", "login"));

    }

    public void internalRolesStuff() throws Exception {
	HybridInternalRoleAdmin admin = ((HybridRealm) realm).getInternalRoleAdmin();
	admin.addInternalRole("ex-admin");
	admin.addInternalRole("ex-dev");

	admin.addUserToInternalRole("deep", "ex-admin");
	admin.addUserToInternalRole("chamith", "ex-admin");
	admin.addUserToInternalRole("deep", "ex-dev");

	// after adding 2 roles
	String[] allRoles = admin.getInternalRoles();
	TestCase.assertEquals(2, allRoles.length);

	String[] roles = admin.getInternalRolesOfUser("deep");
	TestCase.assertEquals(2, roles.length);

	String[] users = admin.getUsersInInternalRole("ex-admin");
	TestCase.assertEquals(2, users.length);

	// removinging one
	admin.removeUserFromInternalRole("deep", "ex-dev");
	roles = admin.getInternalRolesOfUser("deep");
	TestCase.assertEquals(1, roles.length);

    }

    public void createCustomDatabase() throws Exception {

	String users = "CREATE TABLE EMPLOYEES (" + "ID INTEGER GENERATED ALWAYS AS IDENTITY,"
		+ "EMPLOYEE_NAME VARCHAR(255) NOT NULL," + "EMAIL VARCHAR(255) NOT NULL,"
		+ "PRIMARY KEY (ID))";

	String roles = "CREATE TABLE ROLES (" + "ID INTEGER GENERATED ALWAYS AS IDENTITY,"
		+ "ROLE_NAME VARCHAR(255) NOT NULL," + "ROLE_DESCRIPTION VARCHAR(255) NOT NULL,"
		+ "PRIMARY KEY (ID))";

	String user_roles = "CREATE TABLE EMPLOYEE_ROLES ("
		+ "ID INTEGER GENERATED ALWAYS AS IDENTITY," + "ROLE_ID INTEGER NOT NULL,"
		+ "EMP_ID INTEGER NOT NULL,"
		+ "FOREIGN KEY (EMP_ID) REFERENCES EMPLOYEES(ID) ON DELETE CASCADE,"
		+ "FOREIGN KEY (ROLE_ID) REFERENCES ROLES(ID) ON DELETE CASCADE,"
		+ "PRIMARY KEY (ID))";

	Class clazz = Class.forName("org.apache.derby.jdbc.EmbeddedDriver");
	Driver driver = (Driver) clazz.newInstance();
	Properties props = new Properties();
	props.setProperty("user", "tintin");
	props.setProperty("password", "snowy");
	Connection dbConnection = driver.connect("jdbc:derby:target/ComapnyDataBase"
		+ ";create=true", props);

	Statement s = dbConnection.createStatement();
	s.executeUpdate(users);
	s.executeUpdate(roles);
	s.executeUpdate(user_roles);
	dbConnection.commit();

	s
		.executeUpdate("insert into employees(EMPLOYEE_NAME, EMAIL) values ('dimuthul','dimuthul@wso2.com')");
	s
		.executeUpdate("insert into employees(EMPLOYEE_NAME, EMAIL) values ('sameera','sameera@wso2.com')");
	s
		.executeUpdate("insert into employees(EMPLOYEE_NAME, EMAIL) values ('tika','tika@wso2.com')");

	s
		.executeUpdate("insert into roles(ROLE_NAME, ROLE_DESCRIPTION) values ('admin','Admin of comapnay System')");
	s
		.executeUpdate("insert into roles(ROLE_NAME, ROLE_DESCRIPTION) values ('developer','Dev in the company')");

	s.executeUpdate("insert into employee_roles(role_id, emp_id) values (1,1)");
	s.executeUpdate("insert into employee_roles(role_id, emp_id) values (1,2)");
	s.executeUpdate("insert into employee_roles(role_id, emp_id) values (2,2)");
	s.executeUpdate("insert into employee_roles(role_id, emp_id) values (2,3)");

	dbConnection.commit();
	dbConnection.close();

    }

}
