/*
 * Copyright 2004,2005 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.wso2.carbon.utils.deployment;

import org.apache.axiom.om.OMElement;
import org.apache.axis2.AxisFault;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.deployment.*;
import org.apache.axis2.description.AxisService;
import org.apache.axis2.description.AxisServiceGroup;
import org.apache.axis2.engine.AxisConfiguration;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.osgi.framework.Bundle;
import static org.wso2.carbon.utils.WSO2Constants.BUNDLE_ID;

import javax.xml.stream.XMLStreamException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/**
 *
 */
public class ServiceRegistry {

    private static Log log = LogFactory.getLog(ServiceRegistry.class);

    /**
     *
     */
    private ConfigurationContext configCtx;

    /**
     *
     */
    private AxisConfiguration axisConfig;

    /**
     *
     */
    private final Lock lock = new ReentrantLock();

    private Map<Bundle, List<AxisServiceGroup>> serviceGroupMap;

    public ServiceRegistry(ConfigurationContext configCtx) {
        this.configCtx = configCtx;
        this.axisConfig = configCtx.getAxisConfiguration();
        this.serviceGroupMap = new ConcurrentHashMap<Bundle, List<AxisServiceGroup>>();
    }


    /**
     * @param bundle Carbon Compoent that needs to be registred
     */
    public void register(Bundle bundle) {
        lock.lock();
        try {
            addServices(bundle);
        } finally {
            lock.unlock();
        }

    }

    private void addServices(Bundle bundle) {
        if (!serviceGroupMap.containsKey(bundle)) {
            Enumeration enumeration = bundle.findEntries("META-INF", "*services.xml", true);
            List<AxisServiceGroup> axisServiceGroupList = null;
            if (enumeration != null) {
                axisServiceGroupList = new ArrayList<AxisServiceGroup>();
            }
            while (enumeration != null && enumeration.hasMoreElements()) {
                try {
                    ClassLoader loader =
                            new BundleClassLoader(bundle, ServiceRegistry.class.getClassLoader());
                    URL url = (URL) enumeration.nextElement();
                    AxisServiceGroup serviceGroup = new AxisServiceGroup(axisConfig);
                    Dictionary headers = bundle.getHeaders();
                    String bundleSymbolicName = (String) headers.get("Bundle-SymbolicName");
                    serviceGroup.setServiceGroupName(bundleSymbolicName);
                    serviceGroup.addParameter(BUNDLE_ID, bundle.getBundleId());
                    serviceGroup.addParameter("last.updated", bundle.getLastModified());
                    serviceGroup.setServiceGroupClassLoader(loader);
                    InputStream inputStream = url.openStream();
                    DescriptionBuilder builder = new DescriptionBuilder(inputStream, configCtx);
                    OMElement rootElement = builder.buildOM();
                    String elementName = rootElement.getLocalName();
                    HashMap wsdlServicesMap = new HashMap();
                    if (DeploymentConstants.TAG_SERVICE.equals(elementName)) {
                        AxisService axisService = new AxisService(bundleSymbolicName);
                        axisService.setParent(serviceGroup);
                        axisService.setClassLoader(loader);
                        ServiceBuilder serviceBuilder = new ServiceBuilder(configCtx, axisService);
                        serviceBuilder.setWsdlServiceMap(wsdlServicesMap);
                        AxisService service = serviceBuilder.populateService(rootElement);
                        ArrayList serviceList = new ArrayList();
                        serviceList.add(service);
                        DeploymentEngine.addServiceGroup(serviceGroup,
                                serviceList,
                                null,
                                null,
                                axisConfig);
                        if (log.isDebugEnabled()) {
                            log.debug("Deployed axis2 service:" + service.getName() +
                                    " in Bundle: " +
                                    bundle.getSymbolicName());
                        }
                    } else if (DeploymentConstants.TAG_SERVICE_GROUP.equals(elementName)) {
                        ServiceGroupBuilder groupBuilder =
                                new ServiceGroupBuilder(rootElement, wsdlServicesMap,
                                        configCtx);
                        ArrayList serviceList = groupBuilder.populateServiceGroup(serviceGroup);
                        DeploymentEngine.addServiceGroup(serviceGroup,
                                serviceList,
                                null,
                                null,
                                axisConfig);
                        if (log.isDebugEnabled()) {
                            log.debug("Deployed axis2 service group:" +
                                    serviceGroup.getServiceGroupName() + " in Bundle: " +
                                    bundle.getSymbolicName());
                        }
                    }
                    axisServiceGroupList.add(serviceGroup);
                } catch (IOException e) {
                    String msg = "Error building service from bundle : " + bundle.getBundleId();
                    log.error(msg, e);
                } catch (XMLStreamException e) {
                    String msg = "Error reading service.xml from bundle: " + bundle.getBundleId();
                    log.error(msg, e);
                }
            }
            if (axisServiceGroupList != null && axisServiceGroupList.size() > 0) {
                serviceGroupMap.put(bundle, axisServiceGroupList);
            }
        }


    }

    public void unRegister(Bundle bundle) {
        lock.lock();
        try {
            List<AxisServiceGroup> axisServiceGroupList = serviceGroupMap.get(bundle);
            if (axisServiceGroupList != null) {
                for (AxisServiceGroup axisServiceGroup : axisServiceGroupList) {
                    try {
                        axisConfig.removeServiceGroup(axisServiceGroup.getServiceGroupName());
                        if (log.isDebugEnabled()) {
                            log.debug("Stopping" +
                                    axisServiceGroup.getServiceGroupName() +
                                    " service group in Bundle - " +
                                    bundle.getSymbolicName());
                        }
                        for (Iterator iterator = axisServiceGroup.getServices();
                             iterator.hasNext();) {
                            AxisService service = (AxisService) iterator.next();
                            if(log.isDebugEnabled()){
                            log.debug("Service - " + service.getName());
                            }
                        }
                    } catch (AxisFault e) {
                        String msg =
                                "Error while removing services from bundle : " +
                                        bundle.getBundleId();
                        log.error(msg, e);
                    }

                }
                serviceGroupMap.remove(bundle);
            }
        } finally {
            lock.unlock();
        }
    }
}
